/*******************************************************************

Part of the Fritzing project - http://fritzing.org
Copyright (c) 2007-2009 Fachhochschule Potsdam - http://fh-potsdam.de

Fritzing is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Fritzing is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Fritzing.  If not, see <http://www.gnu.org/licenses/>.

********************************************************************

$Revision: 3848 $:
$Author: cohen@irascible.com $:
$Date: 2009-12-10 12:48:23 +0100 (Thu, 10 Dec 2009) $

********************************************************************/


#include <QApplication>
#include <QMouseEvent>

#include "stacktabbar.h"
#include "stacktabwidget.h"
#include "../partsbinpalettewidget.h"
#include "../partsbinview.h"
#include "../../debugdialog.h"


StackTabBar::StackTabBar(StackTabWidget *parent) : QTabBar(parent) {
	setAcceptDrops(true);
	m_parent = parent;
	setProperty("current","false");
	setExpanding(false);
	setElideMode(Qt::ElideRight);
}

int StackTabBar::tabIndexAtPos(const QPoint &p) const {
	for (int i = 0; i < m_parent->count(); ++i) {
    	if (m_parent->isTabEnabled(i) && tabRect(i).contains(p)) {
    		return i;
    	}
	}
    return -1;
}

void StackTabBar::mouseMoveEvent(QMouseEvent *event) {
	// If the left button isn't pressed anymore then return
	if (!(event->buttons() & Qt::LeftButton))
		return;

	// If the distance is too small then return
	if ((event->pos() - m_dragStartPos).manhattanLength()< QApplication::startDragDistance())
		return;

	// initiate Drag
	QDrag* drag = new QDrag(this);

	QMimeData* mimeData = new QMimeData;
	// a crude way to distinguish tab-reodering drops from other ones
	mimeData->setData("action", "tab-reordering");

	drag->setMimeData(mimeData);
	drag->exec();
}

void StackTabBar::mousePressEvent(QMouseEvent *event) {
	if(event->button() == Qt::LeftButton) {
		m_dragStartPos = event->pos();
		int index = tabAt(m_dragStartPos);
		emit setDragSource(m_parent,index);
		if(index == currentIndex()) {
			m_parent->informCurrentChanged(index);
		}
	}

    QTabBar::mousePressEvent(event);
}

void StackTabBar::mouseReleaseEvent(QMouseEvent *event) {
	if(event->button() == Qt::LeftButton) {
		emit setDragSource(NULL);
	}
	QTabBar::mouseReleaseEvent(event);
}

bool StackTabBar::mimeIsAction(const QMimeData* m, const QString& action) {
	if(m) {
		QStringList formats = m->formats();
		return formats.contains("action") && (m->data("action") == action);
	} else {
		return false;
	}
}

void StackTabBar::dragEnterEvent(QDragEnterEvent* event) {
	// Only accept if it's an tab-reordering request
	const QMimeData *m = event->mimeData();
	if (mimeIsAction(m, "tab-reordering") || mimeIsAction(m, "part-reordering")) {
		event->acceptProposedAction();
	}
}

void StackTabBar::dragLeaveEvent(QDragLeaveEvent* event) {
	for(int i=0; i < m_parent->count(); i++) {
		m_parent->showFeedback(i, QTabBar::LeftSide, false);
	}
	QTabBar::dragLeaveEvent(event);
}

void StackTabBar::dragMoveEvent(QDragMoveEvent* event) {
	const QMimeData *m = event->mimeData();
	int index = tabAt(event->pos());
	if(mimeIsAction(m,"tab-reordering")) {
		ButtonPosition side = getButtonPos(index,event->pos());
		emit setPotentialDropSink(m_parent,side,index);
		emit setDropSink(NULL);
		event->acceptProposedAction();
	} else if (event->source() != this && mimeIsAction(event->mimeData(),"part-reordering")) {
		PartsBinPaletteWidget* bin = dynamic_cast<PartsBinPaletteWidget*>(m_parent->widget(index));
		if(bin && !bin->currentBinIsCore()) {
			event->acceptProposedAction();
			setCurrentIndex(index);
		}
	}
}

QTabBar::ButtonPosition StackTabBar::getButtonPos(int index, const QPoint &pos) {
	if(index == count()-1 && posCloserToTheEnd(pos)) {
		return QTabBar::RightSide;
	} else {
		return QTabBar::LeftSide;
	}
}

bool StackTabBar::posCloserToTheEnd(const QPoint &pos) {
	QRect tabRect = this->tabRect(count()-1);
	QPoint middle(tabRect.x()+tabRect.width()/2,tabRect.y()+tabRect.height());
	QRect rightSide(tabRect.topLeft(),middle);
	return !rightSide.contains(pos);
}

void StackTabBar::dropEvent(QDropEvent* event) {
	int toIndex = tabAt(event->pos());

	const QMimeData *m = event->mimeData();
	if(mimeIsAction(m, "tab-reordering")) {
		emit setDropSink(m_parent, getButtonPos(toIndex,event->pos()), toIndex);
		emit dropped();
	} else if(mimeIsAction(m, "part-reordering")) {
		// this widget shouldn't know about bin widgets,
		// but it's already aware of "part-reordering"
		// actions, so fuck it!
		PartsBinPaletteWidget* bin = dynamic_cast<PartsBinPaletteWidget*>(m_parent->widget(toIndex));
		if(bin && !bin->currentBinIsCore()) {
			bin->currentView()->dropEventAux(event,true);
		}
	}

	event->acceptProposedAction();
}

